/**********************************************************************
 *                                                                    *
 *               TP Codeur/Decodeur CIRC / Viterbi                    *
 * bruiteur.c                                                         *
 *                                                                    *
 * Bruiteur : introduction d'erreurs sur un canal                     *
 *   lit un flot d'octets sur l'entree standard et le reecrit sur la  *
 *   sortie standard en introduisant des erreurs.                     *
 *                                                                    *
 *                                                                    *
 * Specification:                                                     *
 *   - Entree: sur l'entree standard, un flot d'octets                *
 *   - Sortie: sur la sortie standard, le meme flot mais avec des     *
 *             introduction d'erreurs.                                *
 *             Le flot de sortie a meme longueur que celui d'entree.  *
 *                                                                    *
 * NB La maniere dont sont introduites les erreurs est controlable    *
 *    par des options: cf procedure "usage" ci-dessous.               *
 *                                                                    *
 *********************************************************************/

void bruiter( double p_err_isole, int l_rafale, double freq_rafale, int deb_rafale ) ;
void usage(char* nomexe) ; 
void LireOptions(int argc, char** argv, 
                 double* p_err_isole, double* freq_rafale, int* l_rafale, int* deb_rafale) ;

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stddef.h>
static int verbose = 0 ;
static int seedalea = 0 ;


/**************************************************/
int main(int argc, char** argv) {
    double p_err_isole  ;
    double freq_rafale ;
    int l_rafale ;
    int deb_rafale ;

    LireOptions ( argc, argv, &p_err_isole, &freq_rafale, &l_rafale, &deb_rafale ) ;
    if (verbose) {
      fprintf(stderr, "*** \n Lancement du bruitage avec les options :\n"
                     "   -p probabilite d'erreur isole            : %f\n" 
                     "   -l longueur des rafales                  : %d\n"
                     "   -f probabilite d'une rafale              : %f\n"
                     "   -d debut d'une rafale forcee             : %d\n"
                     "   -r (1: seed aleatoire / 0: deterministe) : %d\n"
                     "   -v (1: mode verbose   / 0: non verbose ) : %d\n"
	    , p_err_isole, l_rafale, freq_rafale, deb_rafale, seedalea, verbose );
    }  

    bruiter( p_err_isole, l_rafale, freq_rafale, deb_rafale );
    return 0 ;
}


/**************************************************/
void usage( char* nomexe ) {
  fprintf(stderr, 
      "Usage: %s [-h] [-p <f1:flottant>] [-f <f2:flottant>] [-l n3:entier] [-d d4:entier]\n"
      "  Options:\n"
      "    -h : affiche cette aide\n"
      "    -v : verbose\n"
      "    -r : seed arbitraire pour les generations de nombres aleatoires (defaut: seed=0)\n"
      "    -p f1:flottant : probabilite d'erreurs isolee (Valeur defaut=0)\n"
      "    -f f2:flottant : probabilite des rafales d'erreur (Valeur defaut=0)\n" 
      "    -l n3:entier>=0: longeur maximale d'une rafale (Valeur defaut=0)\n" 
      "    -d d4:entier>=0: force une rafale de longueur n3 a partir du\n"
      "                     bit d'indice d4 (defaut=-1: pas de debut de rafale)\n"  
      , nomexe
  ) ;
  fflush(stderr) ;
}


/**************************************************/
void LireOptions(int argc, char** argv, 
                 double* p_err_isole, double* freq_rafale, 
                 int* l_rafale, int* deb_rafale) {
  // on tient compte des arguments
  int i;

  /* Valeurs par defaut */
  *p_err_isole = 0.0 ;
  *freq_rafale = 0.0 ;
  *l_rafale    = 0 ;
  *deb_rafale  = -1;  

  for (i=1; i<argc; i++){
    char* opt =  argv[i] ; 
    if ( (strlen( opt ) != 2) || (opt[0] != '-') )  {
	fprintf( stderr, 
                 "Erreur de format : taper \"bruiteur -h\" pour obtenir de l'aide\n");
	exit(1); 
    } 
    //Cas d'une option
    if (opt[0] == '-'){
      //Y a t il un argument de precise ?
      if ((opt[1] != 'h') && (opt[1] != 'v')  && (opt[1] != 'r')  &&((i+1) == argc)) {
	fprintf( stderr, 
                 "Erreur de format : taper \"bruiteur -h\" pour obtenir de l'aide\n");
	exit(1); 
      } 
      //traitement des options
      switch (opt[1]) {
	case 'h':
	  usage(argv[0]) ; 
	  exit(0) ;
	  break;
	case 'v':
	  verbose = 1;
	  break;
	case 'r':
	  seedalea = 1;
	  break;
	case 'p'://on precise le pourcentage d'erreur isole
	  *p_err_isole = atof(argv[++i]) ; 
	  if (*p_err_isole < 0 || *p_err_isole > 1) {
	    fprintf(stderr, "Erreur de format : "
                            "la probabilite d'erreur isolee doit etre comprise entre 0 et 1.\n");
	    exit(1) ;
	  }
	  break;
	case 'l'://on precise la longueur de la rafale
	  *l_rafale = atoi(argv[++i]) ; 
	  if (*l_rafale < 0) {
	    fprintf(stderr, "Erreur de format : "
	                    "la longueur de rafale doit etre un entier positif.\n") ; 
	    exit(1) ;
	  }
	  break;
	case 'f': //on precise la probabilite des rafales
	  *freq_rafale =  atof(argv[++i]);
	  if (*freq_rafale < 0 || *freq_rafale > 1) {
	    fprintf(stderr, "Erreur de format : "
                            "la probabilite de frequence doit etre comprise entre 0 et 1.\n");
	    exit(1) ;
	  }
	  break;
	case 'd'://taille du fichier
	  *deb_rafale = atoi((char *) argv[++i]);
	  if (*deb_rafale < -1) {
	    fprintf(stderr, "Erreur de format : "
	                    "le debut de rafale doit etre un entier positif ou -1.\n") ; 
	    exit(1) ;
	  }
	  break;
        default: 
	  fprintf(stderr, "Erreur: Option inconnue.\n");
	}
    } 
  }
}




/**************************************************/
void bruiter( double p_err_isolee, int l_rafale, double freq_rafale, int deb_rafale ) {

  char c;

  int rafalecour = 0 ; /* longueur restante de la rafale */
  int nblus = 0 ;
  int nberreurs = 0 ; /* Nombre total d'erreurs introduites */
  int nbraf = 0 ; /* Nombre de rafales */
  int nberrisolees = 0 ; /* Nombre d'erreurs isolees */
  int nberrrafales = 0 ; /* Nombre d'erreurs suite aux rafales */
  int longrafalemax = 0 ; /* Longueur maximale d'une rafale */

  double r = 0;/* probabilite de finir une rafale d'erreur*/
  if (l_rafale==0) {r=0;}
   else { r=1.0/l_rafale;} 

  /* Initialisation du generateur aleatoire: 
   * soit deterministe (seedalea=0); ou sinon arbitraire (option -r)
   */
  srand ( seedalea ? time(NULL) : 0 ) ;

   /* Lecture de l'entree */
   for (nblus=0, c = getchar() ; c != EOF; c=getchar(), nblus++ ) {
     int erreur = 0 ; /* = 1 si on introduit une erreur sur le caractere courant */
     int oldc = c ; 

     /* Priorite aux rafales forcees */
     if (nblus == deb_rafale) {
       erreur = 1 ;
       nberrrafales ++ ; 
       rafalecour = l_rafale-1 ;
       if (rafalecour > longrafalemax ) longrafalemax = rafalecour ;
     }
      
     else /* Priorite a la rafale si une est en cours */
     if (rafalecour >0) { 
       erreur = 1 ;
       nberrrafales ++ ; 
       rafalecour -- ;
     }
     else { /* On teste si une erreur isolee ou une nouvelle rafale est introduite */
       /* Priorite aux rafales */
       if (rand()/(RAND_MAX+1.0) < freq_rafale) { /* Introduction d'une rafale */
         nbraf ++ ; 
         erreur = 1 ;
         nberrrafales ++ ; 
         rafalecour = l_rafale*( rand()/(RAND_MAX+1.0) ) ; 
         if (rafalecour > longrafalemax ) longrafalemax = rafalecour ;
       } 
       else if(rand()/(RAND_MAX+1.0) < p_err_isolee) { /* Introduction d'une erreur isolee */
           erreur = 1 ;
           nberrisolees ++ ; 
       }
     }
     
     if (erreur) c = rand() % 256;
     if (c != oldc) nberreurs++ ;
     putchar(c) ; 
   }

  if (verbose) {
    fprintf(stderr, "Bruitage effectue. Bilan du bruitage :\n"
         "   - nombre d'octets lus           : %d\n" 
         "   - nombre total d'erreurs        : %d\n" 
         "   - nombre d'erreurs isolees      : %d\n" 
         "   - nombre de rafales             : %d\n" 
         "   - longueur maximum d'une rafale : %d\n"  ,
         nblus, nberreurs, nbraf, nberrisolees, nberrrafales, longrafalemax ) ;
  }  
}


